/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

//The following includes must be in this order for CoZ
#include "CoZConstants.h"
#include "DtlLog.h"
#include "sys_util.h"

#include "CoZPacket.h"
#include "CoZClient.h"
#include "CoZUtil.h"

#define Str(x) #x
#define Xstr(x) Str(x)
const char* COZ_VERSION = Xstr(BUILD_VERSION) ;

char* pgm;

/**
 * CoZClient-cmd - command line client for remote execution of commands
 *
 * All of the heavy lifting is done by CozClient class.
 */

void usage() {

#if defined(FROM_DSN)
	fprintf(stderr,
			"usage: fromdsn [OPTION]... //dataset.name\n"
			"   or: fromdsn -ssh [ssh-opt*] [user]@host [OPTION]... //dataset.name\n"
			"   or: fromdsn -local [OPTION]... //dataset.name\n"
			"   or: fromdsn -h\n"
			"   or: fromdsn -v\n"
			"\n"
			"options:\n"
			"  -b  -> \"binary\", disables translation and implies: -k -l none -p 0x00 \n"
			"  -h  -> display this help and exit\n"
			"  -k  -> keep trailing spaces in record; default is to trim spaces for fixed records\n"
			"  -K  -> always trim spaces\n"
			"  -l nl (default) | cr | lf | crlf | crnl | rdw | mfrdw | l4 | 0xbb[bb..] | none \n"
			"     line terminator defaults to nl (newline) except for Windows clients, \n"
			"     which default to crlf. \n"
			"  -L M|A|C|E|W|N|I|D|T|F  -> sets the default logging to the given level: \n"
			"     (eMerg, Alert, Crit, Err, Warning, Info, Notice, Debug, Trace, Fine) \n"
			"  -L logname=level -> changes the level of a specific logname(component) \n"
			"  -L e  -> prefixes log messages with the elapsed time in milliseconds \n"
			"  -L s  -> messages are logged to SYSLOG facility instead of stderr\n"
			"  -L f=filename  -> writes log messages a given file instead of stderr \n"
			"  -o fopen-options\n"
			"  -p 0xbb -> pad character\n"
			"  -q conv_technique\n"
			"  -r  -> \"dataset.name\" is prefixed with MVS userid\n"
			"  -s source_codepage\n"
			"  -t target_codepage\n"
			"  -v  -> display the current version and exit\n"
			"  -x bpxwdyn_alloc_keywords\n"
			"\n"
			"Note: \n"
			"    -ssh -> use an ssh connection to run fromdsn.\n"
			"    -local -> use local z/OS I/O. Useful when CoZLauncher source/target are both z/OS.\n"
			"    If present, -ssh or -local must be the first argument. \n"
			"    If neither -ssh nor -local is not specified, \n"
			"       uses a CoZLauncher socket connection (if configured),  \n"
			"       otherwise, uses a cozcontrol durable session (if configured). \n"
			"    dataset.name is a MVS dataset name or DD:ddname, preceded by '//' \n"
	);

#elif defined(FROM_FILE)
fprintf(stderr,
		"usage: fromfile [OPTION]... filename\n"
		"   or: fromfile -ssh [ssh-opt*] [user]@host [OPTION]... filename\n"
		"   or: fromfile -local [OPTION]... filename\n"
		"   or: fromfile -h\n"
		"   or: fromfile -v\n"
		"\n"
		"options:\n"
		"  -b  -> \"binary\", disables translation \n"
		"  -h  -> display this help and exit\n"
		"  -l nl (default) | cr | lf | crlf | crnl | 0xbb[bb..] | none \n"
		"     line terminator defaults to nl (newline) except for Windows clients, \n"
		"     which default to crlf. \n"
		"  -L M|A|C|E|W|N|I|D|T|F  -> sets the default logging to the given level: \n"
		"     (eMerg, Alert, Crit, Err, Warning, Info, Notice, Debug, Trace, Fine) \n"
		"  -L logname=level -> changes the level of a specific logname(component) \n"
		"  -L e  -> prefixes log messages with the elapsed time in milliseconds \n"
		"  -L s  -> messages are logged to SYSLOG facility instead of stderr\n"
		"  -L f=filename  -> writes log messages a given file instead of stderr \n"
		"  -q conv_technique\n"
		"  -s source_codepage\n"
		"  -t target_codepage\n"
		"  -v  -> display the current version and exit\n"
		"\n"
		"Note: \n"
		"    -ssh -> use an ssh connection to run fromfile.\n"
		"    -local -> use local z/OS I/O. Useful when CoZLauncher source/target are both z/OS.\n"
		"    If present, -ssh or -local must be the first argument. \n"
		"    If neither -ssh nor -local is not specified, \n"
		"       uses a CoZLauncher socket connection (if configured),  \n"
		"       otherwise, uses a cozcontrol durable session (if configured). \n"
		"    filename is a z/OS Unix System Services POSIX path \n"
	);

#elif defined(TO_DSN)
	fprintf(stderr,
			"usage: todsn [OPTION]... //dataset.name\n"
			"   or: todsn -ssh [ssh-opt*] [user]@host [OPTION]... //dataset.name\n"
			"   or: todsn -local [OPTION]... //dataset.name\n"
			"   or: todsn -h\n"
			"   or: todsn -v\n"
			"\n"
			"options:\n"
			"  -a  -> open dataset in append/mod mode\n"
			"  -f  -> begin writing data to the data set immediately (disable buffering)\n"
			"  -b  -> \"binary\", disables translation and implies: -l none -w flow -p 0x00\n"
			"  -h  -> display this help and exit\n"
			"  -l flexible(default) | cr | lf | crlf | nl | crnl | rdw | mfrdw | l4 | 0xbb[bb..] | none \n"
			"     flexible (the default) matches any of: cr, lf, crlf, nl\n"
			"  -L M|A|C|E|W|N|I|D|T|F  -> sets the default logging to the given level: \n"
			"     (eMerg, Alert, Crit, Err, Warning, Info, Notice, Debug, Trace, Fine) \n"
			"  -L logname=level -> changes the level of a specific logname(component) \n"
			"  -L e  -> prefixes log messages with the elapsed time in milliseconds \n"
			"  -L s  -> messages are logged to SYSLOG facility instead of stderr\n"
			"  -L f=filename  -> writes log messages a given file instead of stderr \n"
			"  -o fopen-options\n"
			"  -p 0xbb -> pad character\n"
			"  -q conv_technique\n"
			"  -r  -> \"dataset.name\" is prefixed with MVS userid\n"
			"  -s source_codepage\n"
			"  -t target_codepage\n"
			"  -v  -> display the current version and exit\n"
			"  -w wrap (default) | error | flow | trunc\n"
			"  -x bpxwdyn_alloc_keywords\n"
			"  -z  -> allow for zero length input\n"
			"\n"
			"Note: \n"
			"    -ssh -> use an ssh connection to run todsn.\n"
			"    -local -> use local z/OS I/O. Useful when CoZLauncher source/target are both z/OS.\n"
			"    If present, -ssh or -local must be the first argument. \n"
			"    If neither -ssh nor -local is not specified, \n"
			"       uses a CoZLauncher socket connection (if configured),  \n"
			"       otherwise, uses a cozcontrol durable session (if configured). \n"
			"   dataset.name is either an MVS dataset or DD:ddname, preceded by '//' \n"
	);

#elif defined(TO_FILE)
	fprintf(stderr,
			"usage: tofile [OPTION]... filename\n"
			"   or: tofile -ssh [ssh-opt*] [user]@host [OPTION]... filename\n"
			"   or: tofile -local [OPTION]... filename\n"
			"   or: tofile -h\n"
			"   or: tofile -v\n"
			"\n"
			"options:\n"
			"  -a  -> open file in append mode\n"
			"  -b  -> \"binary\", disables translation\n"
			"  -f  -> begin writing data to the output file immediately (disable buffering)\n"
			"  -h  -> display this help and exit\n"
			"  -l flexible(default) | cr | lf | crlf | nl | crnl | 0xbb[bb..] | none \n"
			"     flexible (the default) matches any of: cr, lf, crlf, nl\n"
			"  -L M|A|C|E|W|N|I|D|T|F  -> sets the default logging to the given level: \n"
			"     (eMerg, Alert, Crit, Err, Warning, Info, Notice, Debug, Trace, Fine) \n"
			"  -L logname=level -> changes the level of a specific logname(component) \n"
			"  -L e  -> prefixes log messages with the elapsed time in milliseconds \n"
			"  -L s  -> messages are logged to SYSLOG facility instead of stderr\n"
			"  -L f=filename  -> writes log messages a given file instead of stderr \n"
			"  -m file access mode to apply to filename (octal)\n"
			"  -n  -> do not replace filename if it exists\n"
			"  -p  -> make path components if they don't exist\n"
			"  -q conv_technique\n"
			"  -s source_codepage\n"
			"  -t target_codepage\n"
			"  -u umask to apply to filename (octal)\n"
			"  -v  -> display the current version and exit\n"
			"  -z  -> allow for zero length input\n"
			"\n"
			"Note: \n"
			"    -ssh -> use an ssh connection to run tofile.\n"
			"    -local -> use local z/OS I/O. Useful when CoZLauncher source/target are both z/OS.\n"
			"    If present, -ssh or -local must be the first argument. \n"
			"    If neither -ssh nor -local is not specified, \n"
			"       uses a CoZLauncher socket connection (if configured),  \n"
			"       otherwise, uses a cozcontrol durable session (if configured). \n"
			"    filename is a z/OS Unix System Services POSIX path \n"
	);

#else
	fprintf(stderr,

"usage: cozclient [OPTION]... command [command-options*]\n"
"   or: cozclient [OPTION]... -ssh [ssh-opt*] [user]@host  command [command-options*]\n"
"   or: cozclient -h\n"
"   or: cozclient -v\n"
"\n"
"options:\n"
"  -i t|b|n -> sets the stdin format (t=text, b=binary,n=none). Default is none. \n"
"  -o t|b|n -> sets the stdout format (t=text, b=binary,n=none). Default is text. \n"
"  -e t|b|s -> sets the stderr format (t=text, b=binary,s=redirect stderr to server). \n"
"              Default is text when -ssh specified or a cozcontrol durable session is used.\n"
"              For CoZLauncher, stderr from the command is sent to Co:Z Server's \n"
"              stderr stream (generally SYSOUT) by default. \n"
"\n"
"Note: \n"
	"    -ssh -> use an ssh connection to run cozclient.\n"
	"    If -ssh is not specified, \n"
	"       uses a CoZLauncher socket connection (if configured),  \n"
	"       otherwise, uses a cozcontrol durable session (if configured). \n"
    "    command is a z/OS command and options.  \n"
);

#endif
}

int main(int argc, char** argv) {


	pgm = basename(argv[0]);

	// Establish the DtlLog program name, using the current base program name
	// add the "-client" suffix for todsn, fromdsn, tofile and fromfile
	char pname[50];
	Stringcopy(pname, pgm, sizeof(pname));
#if defined(TO_DSN) || defined(FROM_DSN) || defined(TO_FILE) || defined(FROM_FILE)
	strncat(pname, "-client", sizeof(pname)-sizeof("-client")-1);
#endif
	SNPRINTF(pname+strlen(pname), sizeof(pname)-strlen(pname), "(%d)", getpid());
	DtlLog::setProgName(pname);
	DtlLog* log = DtlLog::getLogNamed("CoZClient");
	log->printLogName(false);
	log->logmsg(LOG_INFO, "version: %s", COZ_VERSION);

	int i;
	int ssh_opt_start = 0;
	int ssh_opt_num = 0;
	const char* gi = NULL;
	const char* go = NULL;
	const char* ge = NULL;
	const char* gp = NULL;

	const char* cmd_name = NULL;

#if defined(TO_DSN)
	cmd_name = "todsn";
	gi = "b";
	go = "n";
#elif defined(TO_FILE)
	cmd_name = "tofile";
	gi = "b";
	go = "n";
#elif defined(FROM_DSN)
	cmd_name = "fromdsn";
	gi = "n";
	go = "b";
#elif defined(FROM_FILE)
	cmd_name = "fromfile";
	gi = "n";
	go = "b";
#else
	#define FULL_CLIENT 1
	gi = "n";
#endif

	int return_code;
	try {

		for (i=1; i<argc; i++) {
			if (!strcmp(argv[i], "-h")) {
				usage();
	   	 		exit(RC_OK);
			} else if (!strcmp(argv[i], "-v")) {
				printf("%s\n", COZ_VERSION);
    			exit(RC_OK);
			} else if (!strcmp(argv[i], "-ssh")) {
				// Mark the start and number of ssh options
				for (ssh_opt_start=++i; i<argc; i++) {
					if (strchr(argv[i], '@')) {
						ssh_opt_num = i - ssh_opt_start + 1;
						break;
					}
				}
				if (ssh_opt_num) {
					i++;  // point to first actual arg
					break; // done; got ssh options
				} else {
					LOG(LOG_ERR, "-ssh options invalid; must end with user@host");
					throw RC_INPUT_ERROR;
				}
			} else if (!strcmp(argv[i], "-sock")) {
				// sock is the default if not ssh
#ifdef FULL_CLIENT
			} else if (strstr(argv[i], "-i") == argv[i]) {
				if (strlen(argv[i]) > 2) {
					gi = argv[i]+2;
				} else {
					gi = argv[++i];
				}
			} else if (strstr(argv[i], "-o") == argv[i]) {
				if (strlen(argv[i]) > 2) {
					go = argv[i]+2;
				} else {
					go = argv[++i];
				}
			} else if (strstr(argv[i], "-e") == argv[i]) {
				if (strlen(argv[i]) > 2) {
					ge = argv[i]+2;
				} else {
					ge = argv[++i];
				}
#endif
			} else {
				// anything else starts the remote command
				break;
			}
		}

		int arg_offset = i;

		if (argc - arg_offset == 0) {
			usage();
			exit(RC_INPUT_ERROR);
		}

		CoZClient client;

		if (ssh_opt_start) {
			client.setSshOptions(argv, ssh_opt_start, ssh_opt_num);

		} else if (char* p = getenv(ENV_COZ_SSH_OPTS)) {
			// If this environment variable was set, then we use the given
			// options after the ENV_SSH_CMD, and force an ssh type connection,
			// but we ignore if we detect that we are running under CoZAgent or
			// a cozcontrol durable session.

			if (*p && !getenv(ENV_COZ_SERVER_HOST) && !client.selectSocketFileName(true)) {
				char* ssh_opt_string = new char[strlen(p)+1];
				strcpy(ssh_opt_string,p);
				int num_ssh_opts = 50;
				char** ssh_opts = new char*[num_ssh_opts+1];
				CoZUtil::parseCommand(ssh_opt_string, ssh_opts, num_ssh_opts);
				client.setSshOptions(ssh_opts, 0, num_ssh_opts);
				delete [] ssh_opts;  // don't need pointers, but need to keep ssh_opt_string
			}
		}

		if (gi) {
			switch(*gi) {
				case 't':
					client.setIOConfigStdin(IO_CONFIG_TEXT);
					break;
				case 'b':
					client.setIOConfigStdin(IO_CONFIG_BINARY);
					break;
				case 'n':
					client.setIOConfigStdin(IO_CONFIG_NONE);
					break;
				default:
					LOG(LOG_WARNING, "%s: invalid option ignored: -i %s", gi);
			}
		}

		if (go) {
			switch(*go) {
				case 't':
					client.setIOConfigStdout(IO_CONFIG_TEXT);
					break;
				case 'b':
					client.setIOConfigStdout(IO_CONFIG_BINARY);
					break;
				case 'n':
					client.setIOConfigStdout(IO_CONFIG_NONE);
					break;
				default:
					LOG(LOG_WARNING, "%s: invalid option ignored: -o %s", go);
			}
		}

		/* Provide a default for routing stderr depending on whether
		   we are using an ssh connection or not.

		   NOTE: the stderr setting will be forced to text once CoZClient
		         determines that a cozcontrol session will be used.         */

		if (!ge) {
			if (client.isSsh()) {
				ge = "t";  // route stderr as text back to client for ssh connection
			} else {
				ge = "s";  // route stderr directly to server stderr for socket (CoZLauncher)
			}
		} else {
			client.setUserSetStderrOption(true);
		}

		switch(*ge) {
			case 't':
				client.setIOConfigStderr(IO_CONFIG_TEXT);
				break;
			case 'b':
				client.setIOConfigStderr(IO_CONFIG_BINARY);
				break;
			case 's':
				client.setIOConfigStderr(IO_CONFIG_REDIR_TO_SERV);
				break;
			default:
				LOG(LOG_WARNING, "%s: invalid option ignored: -e %s", ge);
		}

#ifdef FULL_CLIENT
		cmd_name = argv[arg_offset++];  // first arg is command name
#endif


		return_code = client.exec(cmd_name, argv, arg_offset, argc-arg_offset);

	} catch (const int rc) {
		return_code = rc;
	} catch (...) {
		LOG(LOG_ERR, "caught unexpectd exception");
    	return_code = RC_SYSTEM_ERROR;
	}

	return return_code;
}



