/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

/**
 * Defines packet protocol for co-zClient <-> co-zServer communications
 */

#ifndef _COZ_PACKET_H
#define _COZ_PACKET_H

static const int  PACKET_HEADER_LEN     = 4;


static const int  MAX_PACKET_PAYLOAD 		= 16 * 1024;
static const int  MAX_CONTROL_PACKET_PAYLOAD = 2 * 1024;

static const int  MAX_PACKET_LEN       		= PACKET_HEADER_LEN + MAX_PACKET_PAYLOAD;
static const int  MAX_CONTROL_PACKET_LEN	= PACKET_HEADER_LEN + MAX_CONTROL_PACKET_PAYLOAD;

static const unsigned char PACKET_TYPE_DATA      = 0;
static const unsigned char PACKET_TYPE_CMD       = 1;

static const unsigned char CMD_CODE_CLIENT_INFO  = 1;
static const unsigned char CMD_CODE_IO_CONFIG  	 = 2;
static const unsigned char CMD_CODE_AUTH_TOKEN	 = 3;
static const unsigned char CMD_CODE_COMMAND		 = 4;
static const unsigned char CMD_CODE_STREAM_BEGINS = 5;
static const unsigned char CMD_CODE_EXIT_CODE 	 = 6; 
static const unsigned char CMD_CODE_ERROR_STATUS = 7;
static const unsigned char CMD_CODE_NOOP 		 = 8;   // Added after 3.6.3/1.2.1
static const unsigned char CMD_CODE_COZCONTROL_ENV_REQ  = 9;
static const unsigned char CMD_CODE_COZCONTROL_ENV      = 10;
static const unsigned char CMD_CODE_COZCONTROL_STOP_REQ = 11;

static const unsigned char IO_CONFIG_BINARY          = 0; 
static const unsigned char IO_CONFIG_TEXT            = 1;
static const unsigned char IO_CONFIG_NONE            = 2; 
static const unsigned char IO_CONFIG_REDIR_TO_SERV   = 3;

static const unsigned char IO_CONFIG_ERRP_NONE       = 0;
static const unsigned char IO_CONFIG_ERRP_INCLUDE    = 1; 

static const unsigned char CLIENT_OPSYS_POSIX        = 1;
static const unsigned char CLIENT_OPSYS_WINDOWS      = 2;

static const int  AUTH_TOKEN_LEN 			= 16;

#pragma pack(1)

struct _CoZPacket {
	unsigned char 	type;
	unsigned char 	code;
	unsigned short 	len;  // non-inclusive of PACKET_HEADER_LEN, network (big endian) byte order
	
	union _payload {
		char raw[MAX_PACKET_PAYLOAD];
		
		struct _client_info { // if code = CMD_CODE_CLIENT_INFO 
			unsigned short version;  // network (big endian) byte order
			unsigned char  client_opsys; // 
			char client_id[255];  // "user@hosname" : null-terminated UTF-8
			// client_codepage immediately follows; null-terminated UTF-8       	   
		} client_info;
		 
		struct _io_config   { // if code = CMD_CODE_IO_CONFIG 
			unsigned char std_in;    // IO_CONFIG_BINARY or IO_CONFIG_TEXT or IO_CONFIG_NONE
			unsigned char std_out;   // IO_CONFIG_BINARY or IO_CONFIG_TEXT or IO_CONFIG_NONE
			unsigned char std_err;   // IO_CONFIG_BINARY or IO_CONFIG_TEXT or IO_CONFIG_REDIR_TO_SERV
			unsigned char errpt;    // IO_CONFIG_ERRP_INCLUDE or IO_CONFIG_ERRP_NONE
		} io_config; 
		
		// if code = CMD_CODE_COZCONTROL_ENV
		struct _env_variable {
			char host[256];
			char port[16];
			char auth_token[2*AUTH_TOKEN_LEN+1]; // full generated auth token
		} env_variable;

		// if code = CMD_CODE_AUTH_TOKEN
		unsigned char auth_token[AUTH_TOKEN_LEN];

		// if code = CMD_CODE_COMAND, then the command string in the client codepage:
		char command[MAX_CONTROL_PACKET_PAYLOAD-1];
		
		// if code = CMD_CODE_EXIT_CODE
		unsigned char exit_code;  // an exit code 0-255
		
		// if code = CMD_CODE_ERROR_STATUS
		// Used to convey encoded error or status information from CoZServer or 
		// a server command (fromdsn, todsn) to a client.  Specifically,
		// fromdsn and todsn will immediately send a "ok" status back with errno=0 if they
		// are able to initialize access to the target dataset. 
		struct _error_status {
			// the following ints are 4-byte network byte order
			int32_t retcode;  // COZ return code
			int32_t errnum;   // system errno
		    int32_t errno2;  //  system errno2
		    uint64_t bytes_received;  // data stream bytes received by server
		    uint64_t bytes_sent;      // data stream bytes sent by server
		    // the following strings are null-terminated in the client code page
		    char component[64];  // component name; null term string in client codepage
		    char msg[128];       // COZ error message
		    char errno_msg[512]; // message from strerror(errno)  
		} error_status;
		
	} payload;
};

#ifdef __MVS__
#pragma pack(reset)
#else
#pragma pack()
#endif

typedef struct _CoZPacket CoZPacket;

#endif  /* _COZ_PACKET_H */
